"""Rename audio files from metadata tags."""

import sys
from importlib import metadata

from .args import fields, parse_args
from .batch import Batch
from .job import Job
from .message import job_info, stats

fields

__version__: str = metadata.version("audiorename")


def execute(*argv: str) -> None:
    """Main function

    :param list argv: The command line arguments specified as a list: e. g
        :code:`['--dry-run', '.']`
    """

    job = None

    try:
        args = parse_args(argv)
        job = Job(args)
        job.stats.counter.reset()
        job.stats.timer.start()
        if job.cli_output.job_info:
            job_info(job)
        if job.rename.dry_run:
            job.msg.output("Dry run")
        batch = Batch(job)
        batch.execute()
        job.stats.timer.stop()
        if job.cli_output.stats:
            stats(job)
    except KeyboardInterrupt:
        if job:
            job.stats.timer.stop()
            if job.cli_output.stats:
                stats(job)
        sys.exit(0)



"""Batch processing of the audio files."""

import os
import typing

from phrydy.mediafile_extended import MediaFileExtended

from .audiofile import do_job_on_audiofile, mb_track_listing
from .job import Job


class VirtualAlbum:
    title: str
    track: int
    path: str

    def __init__(self, title: str, track: int, path: str) -> None:
        self.title = title
        self.track = track
        self.path = path


class Batch:
    """This class first sorts all files and then walks through all files. In
    this process it tries to make bundles of files belonging to an album. This
    bundle of files is temporary stored in the attribute `virtual_album`. This
    complicated mechanism is needed for the two filters `album_complete` and
    `album_min`.
    """

    virtual_album: typing.List[VirtualAlbum] = []
    """Storage of a list of files belonging to an album."""

    current_album_title: str = ""
    """Storage for the album title of the current audio file."""

    job: Job

    bundle_filter: bool

    def __init__(self, job: Job):
        self.job = job
        self.bundle_filter = job.filters.album_complete or isinstance(
            job.filters.album_min, int
        )

    def check_extension(self, path: str) -> bool:
        """Check the extension of the track.

        :params str path: The path of the tracks.
        """
        extension = self.job.filters.extension
        extension = ["." + e for e in extension]
        if path.lower().endswith(tuple(extension)):
            return True
        else:
            return False

    def check_quantity(self):
        """Compare the number of tracks in an album with the minimal track
        threshold.
        """
        if isinstance(self.job.filters.album_min, int) and len(
            self.virtual_album
        ) > int(self.job.filters.album_min):
            return True
        else:
            return False

    def check_completeness(self):
        """Check if the album is complete"""
        max_track = 0
        for album in self.virtual_album:
            if album.track > max_track:
                max_track = album.track

        if len(self.virtual_album) == max_track:
            return True
        else:
            return False

    def process_album(self):
        """Check an album for quantity and completeness."""
        quantity = True
        completeness = True
        if self.job.filters.album_min and not self.check_quantity():
            quantity = False
        if self.job.filters.album_complete and not self.check_completeness():
            completeness = False

        if quantity and completeness:
            for album in self.virtual_album:
                do_job_on_audiofile(album.path, job=self.job)

        self.virtual_album = []

    def make_bundles(self, path: str = ""):
        """
        :params str path: The path of the tracks.
        """
        if not path:
            self.process_album()
            return

        try:
            media = MediaFileExtended(path)
            album = VirtualAlbum(media.album, media.track, path)
            if not self.current_album_title or self.current_album_title != media.album:
                self.current_album_title = media.album
                self.process_album()
            self.virtual_album.append(album)
        except Exception:
            pass

    def execute(self):
        """Process all files of a given path or process a single file."""

        mb_track_listing.counter = 0
        if os.path.isdir(self.job.selection.source):
            for path, dirs, files in os.walk(self.job.selection.source):
                dirs.sort()
                files.sort()
                for file_name in files:
                    p = os.path.join(path, file_name)
                    if self.check_extension(p):
                        if self.bundle_filter:
                            self.make_bundles(p)
                        else:
                            do_job_on_audiofile(p, job=self.job)

            # Process the last bundle left over
            if self.bundle_filter:
                self.make_bundles()

        else:
            p = self.job.selection.source
            if self.check_extension(p):
                do_job_on_audiofile(p, job=self.job)



"""Create the command line interface using the package “argparse”."""

import argparse
import os
from typing import Any, List, Optional, Tuple, Union, cast

import phrydy
import tmep
from typing_extensions import Literal

from .utils import indent, read_file

fields: phrydy.field_docs.FieldDocCollection = {
    "ar_classical_album": {
        "description": "The field “work” without the movement suffix. "
        "For example: “Horn Concerto: I. Allegro” -> "
        "“Horn Concerto”",
        "category": "common",
        "examples": ["Horn Concerto", "Die Meistersinger von Nürnberg"],
    },
    "ar_classical_performer": {
        "description": "“ar_performer_short” or “albumartist” without the "
        "composer prefix: “Beethoven; Karajan, Mutter” -> "
        "“Karajan, Mutter”",
        "category": "common",
        "examples": ["Karajan, Mutter", "Karajan, StaDre"],
        "data_type": "str",
    },
    "ar_classical_title": {
        "description": "The movement title without the parent work prefix. "
        "For example “Horn Concerto: I. Allegro” -> "
        "“I. Allegro”",
        "category": "common",
        "examples": [
            "I. Allegro",
            'Akt III, Szene V. "Morgendlich leuchtend im rosigen '
            'Schein" (Walther, Volk, Meister, Sachs, Pogner, Eva)',
        ],
        "data_type": "str",
    },
    "ar_classical_track": {
        "description": "If the title contains Roman numbers, then these are "
        "converted to arabic numbers with leading zeros. "
        "If no Roman numbers could be found, then the field "
        "“ar_combined_disctrack” is used.",
        "category": "common",
        "examples": ["01", "4-08"],
        "data_type": "str",
    },
    "ar_combined_album": {
        "description": "“album” without ” (Disc X)”.",
        "category": "common",
        "examples": [
            "Headlines and Deadlines: The Hits of a-ha",
            "Die Meistersinger von Nürnberg",
        ],
    },
    "ar_combined_artist": {
        "description": "The first non-empty value of the following list of "
        "fields: "
        "“albumartist” -> “artist” -> “albumartist_credit” "
        "-> “artist_credit” -> “albumartist_sort” -> "
        "“artist_sort”. "
        "If no value could be determined, then “Unknown” is "
        "assigned. "
        "The second artist after “feat.”, “ft.” or “vs.” "
        "is removed.",
        "category": "common",
        "examples": ["a-ha", "Richard Wagner; René Kollo, Helen Donath, ..."],
        "data_type": "str",
    },
    "ar_combined_artist_sort": {
        "description": "The first non-empty value of the following list of "
        "fields: "
        "“albumartist_sort” -> “artist_sort” -> "
        "“albumartist” -> “artist” -> “albumartist_credit” -> "
        "“artist_credit”. "
        "If no value could be determined, then “Unknown” is "
        "assigned. "
        "The second artist after “feat.”, “ft.” or “vs.” "
        "is removed.",
        "category": "common",
        "examples": ["a-ha", "Wagner, Richard; Kollo, René, Donath, Helen..."],
        "data_type": "str",
    },
    "ar_combined_composer": {
        "description": "The first not empty field of this field list: "
        "“composer_sort”, “composer”, “ar_combined_artist”",
        "category": "common",
        "examples": ["Beethoven, Ludwig-van", "Wagner, Richard"],
        "data_type": "str",
    },
    "ar_combined_disctrack": {
        "description": "Combination of disc and track in the format: " "disk-track",
        "category": "common",
        "examples": ["1-01", "3-099"],
        "data_type": "str",
    },
    "ar_combined_soundtrack": {
        "description": "Boolean flag which indicates if the audio file is "
        "a soundtrack",
        "category": "common",
        "examples": [True, False],
        "data_type": "bool",
    },
    "ar_combined_work_top": {
        "description": "The work on the top level of a work hierarchy.",
        "category": "common",
        "examples": ["Horn Concerto: I. Allegro", "Die Meistersinger von Nürnberg"],
        "data_type": "str",
    },
    "ar_combined_year": {
        "description": "First “original_year” then “year”.",
        "category": "common",
        "examples": [1978],
        "data_type": "int",
    },
    "ar_initial_album": {
        "description": "First character in lowercase of “ar_combined_album”. "
        "Allowed characters: [a-z, 0, _], 0-9 -> 0, ? -> _. "
        "For example “Help!” -> “h”.",
        "category": "common",
        "examples": ["h"],
    },
    "ar_initial_artist": {
        "description": "First character in lowercase of "
        "“ar_combined_artist_sort”. "
        "Allowed characters: [a-z, 0, _], 0-9 -> 0, ? -> _. "
        "For example “Brendel, Alfred” -> “b”.",
        "category": "common",
        "examples": ["b"],
        "data_type": "str",
    },
    "ar_initial_composer": {
        "description": "First character in lowercase of "
        "“ar_combined_composer”. "
        "Allowed characters: [a-z, 0, _], 0-9 -> 0, ? -> _. "
        "For example “Ludwig van Beethoven” -> “l”.",
        "category": "common",
        "examples": ["l"],
        "data_type": "str",
    },
    "ar_performer": {
        "description": "Performer names.",
        "category": "common",
        "examples": ["Herbert von Karajan, Staatskapelle Dresden"],
        "data_type": "str",
    },
    "ar_performer_raw": {
        "description": "Raw performer names.",
        "category": "common",
        "examples": [
            [
                ["conductor", "Herbert von Karajan"],
                ["orchestra", "Staatskapelle Dresden"],
            ]
        ],
        "data_type": "list",
    },
    "ar_performer_short": {
        "description": "Abbreviated performer names.",
        "category": "common",
        "examples": ["Karajan, StaDre"],
        "data_type": "str",
    },
}
"""Documentation of the extra fields."""

all_fields = phrydy.doc_generator.merge_fields(phrydy.field_docs.fields, fields)


class ArgsDefault:
    """To document the return value of the :func:`audiorename.args.parse_args`.
    It can also be used to mock the args object for testing purposes.
    """

    # default
    config: Optional[List[str]] = None

    # [selection]
    source: Optional[str] = None
    source_as_target: Optional[bool] = None
    target: Optional[str] = None

    # [rename]
    backup_folder: Optional[str] = None
    best_format: Optional[bool] = None
    dry_run: Optional[bool] = None
    move_action: Union[Literal["move", "copy", "no_rename"], None] = None
    cleaning_action: Union[Literal["backup", "delete", "do_nothing"], None] = None

    # [filters]
    album_complete: Optional[bool] = None
    album_min: Optional[int] = None
    extension: Optional[str] = None
    genre_classical: Optional[str] = None
    field_skip: Optional[str] = None

    # [template_settings]
    classical: Optional[bool] = None
    shell_friendly: Optional[bool] = None
    no_soundtrack: Optional[bool] = None

    # [path_templates]
    default_template: Optional[str] = None
    soundtrack_template: Optional[str] = None
    compilation_template: Optional[str] = None
    classical_template: Optional[str] = None

    # [cli_output]
    color: Optional[bool] = None
    debug: Optional[bool] = None
    job_info: Optional[bool] = None
    mb_track_listing: Optional[bool] = None
    one_line: Optional[bool] = None
    stats: Optional[bool] = None
    verbose: Optional[bool] = None

    # [metadata_actions]
    enrich_metadata: Optional[bool] = None
    remap_classical: Optional[bool] = None

    def __init__(self, **kwargs: Any):
        for k, v in kwargs.items():
            setattr(self, k, v)


def read_configuration_file() -> str:
    content = read_file(
        os.path.join(os.path.dirname(os.path.abspath(__file__)), "example-config.ini")
    )
    return indent(content)


def description() -> str:
    """Build the description string."""
    return (
        """\
    Rename audio files from metadata tags.

    How to specify the target directory?

    1. By the default the audio files are moved or renamed to the parent
       working directory.
    2. Use the option ``-t <folder>`` or ``--target <folder>`` to specifiy
       a target directory.
    3. Use the option ``-a`` or ``--source-as-target`` to copy or rename
       your audio files within the source directory.

Metadata fields
===============

"""
        + phrydy.format_fields_as_txt(
            additional_fields=fields, color=True, field_prefix="$"
        )
        + """

Functions
=========

"""
        + tmep.doc.Doc().get()
        + """

Configuration file
==================

"""
        + read_configuration_file()
    )


def parse_args(argv: Optional[Tuple[str]]) -> ArgsDefault:
    """Parse the command line arguments using the python library `argparse`.

    :param list argv: The command line arguments specified as a list: e. g
        :code:`['--dry-run', '.']`

    :return: Dictionary see :class:`audiorename.args.ArgsDefault`
    """

    if not argv:
        argv = None

    parser = argparse.ArgumentParser(
        formatter_class=argparse.RawDescriptionHelpFormatter, description=description()
    )

    ##
    # Options without section. These options have no equivalent in the
    # configuration file
    ##

    # config
    parser.add_argument(
        "--config",
        help="Load a configuration file in INI format.",
        action="append",
        default=None,
    )

    # version
    parser.add_argument(
        "-v",
        "--version",
        action="version",
        version="%(prog)s {version}".format(version="0.0.0"),
    )

    ###############################################################################
    # [selection]
    ###############################################################################

    selection = parser.add_argument_group(
        title="[selection]",
        description="The following arguments are intended to select the audio "
        "files.",
    )

    # source
    selection.add_argument(
        "source",
        help="A folder containing audio files or a single audio file. If you "
        "specify a folder, the program will search for audio files in all "
        "subfolders. If you want to rename the audio files in the current "
        "working directory, then specify a dot (“.”).",
    )

    # target
    selection.add_argument(
        "-t",
        "--target",
        help="Target directory",
        default=None,
    )

    # source_as_target
    selection.add_argument(
        "-a",
        "--source-as-target",
        help="Use specified source folder as target directory",
        action="store_true",
        default=None,
    )

    ###############################################################################
    # Rename
    ###############################################################################

    rename = parser.add_argument_group(
        title="[rename]",
        description="These options configure the actual renaming process.",
    )

    # backup_folder
    rename.add_argument(
        "-p",
        "--backup-folder",
        help="Folder to store the backup files in.",
        default=None,
    )

    # best_format
    rename.add_argument(
        "-B",
        "--best-format",
        help="Use the best format. This option only takes effect if the \
        target file already exists. `audiorename` now checks the qualtity of \
        the two audio files (source and target). The tool first examines the \
        format. For example a FLAC file wins over a MP3 file. Then \
        `audiorename` checks the bitrate.",
        action="store_true",
        default=None,
    )

    # dry_run
    rename.add_argument(
        "-d",
        "--dry-run",
        help="Don’t rename or copy the audio files.",
        action="store_true",
        default=None,
    )

    ##
    # Move actions
    ##

    rename_move = parser.add_argument_group("move action")
    exclusive_rename_move = rename_move.add_mutually_exclusive_group()

    # move_action: copy
    exclusive_rename_move.add_argument(
        "-C",
        "--copy",
        dest="move_action",
        help="Copy files instead of rename / move.",
        action="store_const",
        const="copy",
    )

    # move_action: move
    exclusive_rename_move.add_argument(
        "-M",
        "--move",
        dest="move_action",
        help="Move / rename a file. This is the default action. The option \
        can be omitted.",
        action="store_const",
        const="move",
    )

    # move_action: no_rename
    exclusive_rename_move.add_argument(
        "-n",
        "--no-rename",
        dest="move_action",
        help="Don’t rename, move, copy or perform a dry run. Do nothing.",
        action="store_const",
        const="no_rename",
    )

    ##
    # Cleaning actions
    ##

    rename_cleaning = parser.add_argument_group(
        title="cleaning action",
        description="The cleaning actions are only executed if the target "
        "file already exists.",
    )
    exclusive_rename_cleaning = rename_cleaning.add_mutually_exclusive_group()

    # cleaning_action: backup
    exclusive_rename_cleaning.add_argument(
        "-A",
        "--backup",
        dest="cleaning_action",
        help="Backup the audio files instead of deleting them. The backup \
        directory can be specified with the --backup-folder option.",
        action="store_const",
        const="backup",
    )

    # cleaning_action: delete
    exclusive_rename_cleaning.add_argument(
        "-D",
        "--delete",
        dest="cleaning_action",
        help="Delete the audio files instead of creating a backup.",
        action="store_const",
        const="delete",
    )

    ###############################################################################
    # filters
    ###############################################################################

    filters = parser.add_argument_group(
        title="[filters]",
        description="The following options filter the music files that are "
        "renamed according to certain rules.",
    )

    # album_complete
    filters.add_argument(
        "-F",
        "--album-complete",
        help="Rename only complete albums.",
        action="store_true",
        default=None,
    )

    # album_min
    filters.add_argument(
        "-m",
        "--album-min",
        type=int,
        help="Rename only albums containing at least X files.",
        default=None,
    )

    # extension
    filters.add_argument(
        "-e", "--extension", help="Extensions to rename.", default=None
    )

    # genre classical
    filters.add_argument(
        "--genre-classical", help="List of genres to be classical.", default=","
    )

    # field_skip
    filters.add_argument(
        "-s", "--field-skip", help="Skip renaming if field is empty.", default=None
    )

    ###############################################################################
    # formats
    ###############################################################################

    template_settings = parser.add_argument_group("[template_settings]")

    # classical
    template_settings.add_argument(
        "-k",
        "--classical",
        help="Use the default format for classical music. If you use this \
        option, both parameters (--default and --compilation) have no \
        effect. Classical music is sorted by the lastname of the composer.",
        action="store_true",
        default=None,
    )

    # shell_friendly
    template_settings.add_argument(
        "-S",
        "--shell-friendly",
        help="Rename audio files “shell friendly”, this means without \
        whitespaces, parentheses etc.",
        action="store_true",
        default=None,
    )

    # no_soundtrack
    template_settings.add_argument(
        "--no-soundtrack",
        action="store_true",
        help="Do not use the path template for soundtracks. Use instead the \
        default path template.",
        default=None,
    )

    ###############################################################################
    # path_templates
    ###############################################################################

    path_templates = parser.add_argument_group(
        title="[path_templates]",
        description="audiorename provides default path templates. "
        "You can specify your own path templates using the following options.",
    )

    # default_template
    path_templates.add_argument(
        "-f",
        "--default",
        "--format",
        metavar="PATH_TEMPLATE",
        dest="default_template",
        help="The default path template for audio files that are not \
        compilations or compilations. Use metadata fields and functions to \
        build the path template.",
        default=None,
    )

    # compilation_template
    path_templates.add_argument(
        "-c",
        "--compilation",
        metavar="PATH_TEMPLATE",
        dest="compilation_template",
        help="Path template for compilations. Use metadata fields and \
        functions to build the path template.",
        default=None,
    )

    # soundtrack_template
    path_templates.add_argument(
        "--soundtrack",
        metavar="PATH_TEMPLATE",
        dest="soundtrack_template",
        help="Path template for a soundtrack audio file. Use metadata fields \
        and functions to build the path template.",
        default=None,
    )

    # classical_template
    path_templates.add_argument(
        "--format-classical",
        metavar="PATH_TEMPLATE",
        dest="classical_template",
        help="Path template for classical audio file. Use metadata fields \
        and functions to build the path template.",
        default=None,
    )

    ###############################################################################
    # cli_output
    ###############################################################################

    cli_output = parser.add_argument_group(
        title="[cli_output]",
        description="This group contains all options that affect the output "
        "on the command line interface (cli).",
    )

    output_color = cli_output.add_mutually_exclusive_group()

    # color
    output_color.add_argument(
        "-K",
        "--color",
        help="Colorize the standard output of the program with ANSI colors.",
        action="store_true",
        default=None,
    )

    output_color.add_argument(
        "--no-color",
        help="Don’t colorize the standard output of the program with ANSI " "colors.",
        action="store_false",
        dest="color",
        default=None,
    )

    # debug
    cli_output.add_argument(
        "-b",
        "--debug",
        help="Print debug informations about the single metadata fields.",
        action="store_true",
        default=None,
    )

    # job_info
    cli_output.add_argument(
        "-j",
        "--job-info",
        help="Display informations about the current job. This informations \
        are printted out before any actions on the audio files are executed.",
        action="store_true",
        default=None,
    )

    # mb_track_listing
    cli_output.add_argument(
        "-l",
        "--mb-track-listing",
        help="Print track listing for Musicbrainz website: Format: track. \
        title (duration), e. g.: \
          1. He, Zigeuner (1:31) \
          2. Hochgetürmte Rimaflut (1:21)",
        action="store_true",
        default=None,
    )

    # one_line
    cli_output.add_argument(
        "-o",
        "--one-line",
        help="Display the rename / copy action status on one line instead of \
        two.",
        action="store_true",
        default=None,
    )

    # stats
    cli_output.add_argument(
        "-T",
        "--stats",
        help="Show statistics at the end of the execution.",
        action="store_true",
        default=None,
    )

    # verbose
    cli_output.add_argument(
        "-V",
        "--verbose",
        help="Make the command line output more verbose.",
        action="store_true",
        default=None,
    )

    ###############################################################################
    # Metadata actions
    ###############################################################################

    metadata_actions = parser.add_argument_group("[metadata_actions]")

    # enrich_metadata
    metadata_actions.add_argument(
        "-E",
        "--enrich-metadata",
        help="Fetch the tag fields “work” and “mb_workid” from Musicbrainz \
        and save this fields into the audio file. The audio file must have \
        the tag field “mb_trackid”. The give audio file is not renamed.",
        action="store_true",
        default=None,
    )

    # remap_classical
    metadata_actions.add_argument(
        "-r",
        "--remap-classical",
        help="Remap some fields to fit better for classical music: \
        “composer” becomes “artist”, “work” becomes “album”, from the \
        “title” the work prefix is removed (“Symphonie No. 9: I. Allegro” \
        -> “I. Allegro”) and “track” becomes the movement number. All \
        overwritten fields are safed in the “comments” field.",
        action="store_true",
        default=None,
    )

    return cast(ArgsDefault, parser.parse_args(argv))


def format_fields_as_rst_table() -> str:
    return phrydy.doc_generator.format_fields_as_rst_table(additional_fields=fields)



"""This module contains all functionality on the level of a single audio file."""

import errno
import os
import re
import shutil
import traceback
from typing import Any, Dict, List, Literal, Optional

import phrydy
from tmep import Functions, Template
from tmep.format import asciify, delchars, deldupchars, replchars

from .job import Job
from .meta import Meta, compare_dicts

DestinationType = Literal["source", "target"]


class AudioFile:
    """
    :param path: The path string of the audio file.
    :param job: The current `job` object.
    :param string file_type: Either “source” or “target”.
    :param string prefix: The path prefix of the audio file, for example the
        base folder of your music collection. Used to shorten the path strings
        in the progress messaging.
    """

    __path: str

    type: DestinationType
    job: Job
    __prefix: Optional[str]

    def __init__(
        self,
        path: str,
        job: Job,
        file_type: DestinationType = "source",
        prefix: Optional[str] = None,
    ):
        self.__path = path
        self.type = file_type
        self.job = job
        self.__prefix = prefix
        self.shorten_symbol = "[…]"

    @property
    def shell_friendly(self):
        if not self.job:
            return True
        else:
            return self.job.template_settings.shell_friendly

    @property
    def meta(self) -> Optional[Meta]:
        if self.exists:
            try:
                return Meta(self.abspath, self.shell_friendly)
            except Exception as e:
                tb = traceback.TracebackException.from_exception(e)
                print("".join(tb.stack.format()))
        return None

    @property
    def abspath(self) -> str:
        """The absolute path of the audio file."""
        return os.path.abspath(self.__path)

    @property
    def prefix(self):
        if self.__prefix and len(self.__prefix) > 1:
            if self.__prefix[-1] != os.path.sep:
                return self.__prefix + os.path.sep
            else:
                return self.__prefix

    @property
    def exists(self):
        return os.path.exists(self.abspath)

    @property
    def extension(self) -> str:
        """The file extension of the audio file."""
        return self.abspath.split(".")[-1].lower()

    @property
    def short(self) -> str:
        if self.prefix:
            short = self.abspath.replace(self.prefix, "")
        else:
            short = os.path.basename(self.abspath)

        return self.shorten_symbol + short

    @property
    def filename(self) -> str:
        """The file name of the audio file."""
        return os.path.basename(self.abspath)

    @property
    def dir_and_file(self) -> str:
        """The parent directory name and the file name."""
        path_segments = self.abspath.split(os.path.sep)
        return os.path.sep.join(path_segments[-2:])


class MBTrackListing:
    def __init__(self):
        self.counter = 0

    def format_audiofile(self, album: str, title: str, length: int) -> str:
        self.counter += 1

        m, s = divmod(length, 60)
        mmss = "{:d}:{:02d}".format(int(m), int(s))
        output = "{:d}. {:s}: {:s} ({:s})".format(self.counter, album, title, mmss)
        output = output.replace("Op.", "op.")
        return output.replace("- ", "")


mb_track_listing = MBTrackListing()


def find_target_path(target: str, extensions: List[str]) -> Optional[str]:
    """Get the path of a existing audio file target. Search for audio files
    with different extensions.
    """
    target = os.path.splitext(target)[0]
    for extension in extensions:
        audio_file = target + "." + extension
        if os.path.exists(audio_file):
            return audio_file
    return None


def detect_best_format(source: Meta, target: Meta, job: Job) -> DestinationType:
    """
    :param source: The metadata object of the source file.
    :param target: The metadata object of the target file.
    :param job: The `job` object.

    :return: Either the string `source` or the string `target`
    """

    def get_highest(dictionary: Dict[Any, DestinationType]) -> DestinationType:
        out: DestinationType = "target"
        for _, value in sorted(dictionary.items()):
            out = value
        return out

    if source.format == target.format:
        bitrates: Dict[int, Literal["source", "target"]] = {}
        bitrates[source.bitrate] = "source"
        bitrates[target.bitrate] = "target"
        best = get_highest(bitrates)
        job.msg.best_format(best, "bitrate", source, target)
        return best

    else:
        # All types:
        #
        # 'aac'
        # 'aiff'
        # 'alac': Apple Lossless Audio Codec (losless)
        # 'ape'
        # 'asf'
        # 'dsf'
        # 'flac'
        # 'mp3'
        # 'mpc'
        # 'ogg'
        # 'opus'
        # 'wv': WavPack (losless)

        ranking = {
            "flac": 10,
            "alac": 9,
            "aac": 8,
            "mp3": 5,
            "ogg": 2,
            "wma": 1,
        }

        types = {}
        types[ranking[source.type] if source.type in ranking else 0] = "source"
        types[ranking[target.type] if target.type in ranking else 0] = "target"
        best = get_highest(types)
        job.msg.best_format(best, "type", source, target)
        return best


def process_target_path(meta: Meta, format_string: str, shell_friendly: bool = True):
    """
    :param dict meta: The to a dictionary converted attributes of a
        meta object :class:`audiorename.meta.Meta`.
    :param string format_string:
    :param boolean shell_friendly:
    """
    template = Template(format_string)
    functions = Functions(meta)
    target = template.substitute(meta, functions.functions())

    if isinstance(target, str):
        if shell_friendly:
            target = asciify(target)
            target = delchars(target, "().,!\"'’")
            target = replchars(target, "-", " ")
        # asciify generates new characters which must be sanitzed, e. g.:
        # ¿ -> ?
        target = delchars(target, ':*?"<>|\\~&{}')
        target = deldupchars(target)
    return re.sub(r"\.$", "", target)


class Action:
    """
    :param job: The `job` object.
    :type job: audiorename.job.Job
    """

    job: Job

    def __init__(self, job: Job):
        self.job = job
        self.dry_run = job.rename.dry_run

    def count(self, counter_name: str):
        self.job.stats.counter.count(counter_name)

    def cleanup(self, audio_file: AudioFile):
        if self.job.rename.cleaning_action == "backup":
            self.backup(audio_file)
        elif self.job.rename.cleaning_action == "delete":
            self.delete(audio_file)

    def backup(self, audio_file: AudioFile):
        backup_file = AudioFile(
            os.path.join(
                self.job.rename.backup_folder, os.path.basename(audio_file.abspath)
            ),
            job=self.job,
            file_type="target",
        )

        self.job.msg.action_two_path("Backup", audio_file, backup_file)
        self.count("backup")
        if not self.dry_run:
            self.create_dir(backup_file)
            shutil.move(audio_file.abspath, backup_file.abspath)

    def copy(self, source: AudioFile, target: AudioFile):
        self.job.msg.action_two_path("Copy", source, target)
        self.count("copy")
        if not self.dry_run:
            self.create_dir(target)
            shutil.copy2(source.abspath, target.abspath)

    def create_dir(self, audio_file: AudioFile):
        path = os.path.dirname(audio_file.abspath)

        try:
            os.makedirs(path)
        except OSError as exception:
            if exception.errno != errno.EEXIST:
                raise

    def delete(self, audio_file: AudioFile):
        self.job.msg.action_one_path("Delete", audio_file)
        self.count("delete")
        if not self.dry_run:
            os.remove(audio_file.abspath)

    def move(self, source: AudioFile, target: AudioFile):
        self.job.msg.action_two_path("Move", source, target)
        self.count("move")
        if not self.dry_run:
            self.create_dir(target)
            shutil.move(source.abspath, target.abspath)

    def metadata(
        self, audio_file: AudioFile, enrich: bool = False, remap: bool = False
    ) -> None:
        if not audio_file.meta:
            raise Exception("The given audio file has no meta property.")
        meta = audio_file.meta
        pre = meta.export_dict(sanitize=False)

        def single_action(
            meta: Meta,
            method_name: Literal["enrich_metadata", "remap_classical"],
            message: str,
        ):
            pre = meta.export_dict(sanitize=False)
            method = getattr(meta, method_name)
            method()
            post = meta.export_dict(sanitize=False)
            diff = compare_dicts(pre, post)
            if diff:
                self.count(method_name)
            self.job.msg.output(message)
            for change in diff:
                self.job.msg.diff(change[0], change[1], change[2])

        if enrich:
            single_action(meta, "enrich_metadata", "Enrich metadata")
        if remap:
            single_action(meta, "remap_classical", "Remap classical")

        post = meta.export_dict(sanitize=False)
        diff = compare_dicts(pre, post)

        if not self.dry_run and diff:
            meta.save()


def do_job_on_audiofile(source_path: str, job: Job):
    def count(key: str):
        job.stats.counter.count(key)

    skip = False

    action = Action(job)

    source = AudioFile(source_path, job=job, prefix=os.getcwd(), file_type="source")
    if not job.cli_output.mb_track_listing:
        job.msg.next_file(source)

    if not source.meta:
        skip = True

    ##
    # Skips
    ##

    if skip:
        job.msg.status("Broken file", status="error")
        count("broken_file")
        return

    ##
    # Output only
    ##

    if not source.meta:
        raise Exception("source.meta must not be empty.")

    if job.cli_output.mb_track_listing:
        print(
            mb_track_listing.format_audiofile(
                source.meta.album, source.meta.title, source.meta.length
            )
        )
        return

    if job.cli_output.debug:
        phrydy.print_debug(
            source.abspath,
            Meta,
            Meta.fields,
            job.cli_output.color,
        )
        return

    if job.filters.field_skip and (
        not hasattr(source.meta, job.filters.field_skip)
        or not getattr(source.meta, job.filters.field_skip)
    ):
        job.msg.status("No field", status="error")
        count("no_field")
        return

    ##
    # Metadata actions
    ##

    if job.metadata_actions.remap_classical or job.metadata_actions.enrich_metadata:
        action.metadata(
            source,
            job.metadata_actions.enrich_metadata,
            job.metadata_actions.remap_classical,
        )

    if (
        source.meta.genre is not None
        and getattr(source.meta, "genre", "").lower() in job.filters.genre_classical
    ):
        if not job.metadata_actions.remap_classical:
            action.metadata(source, job.metadata_actions.enrich_metadata, True)

    ##
    # Rename action
    ##

    if job.rename.move_action != "no_rename":
        if (
            source.meta.genre is not None
            and getattr(source.meta, "genre", "").lower() in job.filters.genre_classical
        ):
            format_string = job.path_templates.classical
        elif source.meta.ar_combined_soundtrack:
            if job.args.no_soundtrack and source.meta.comp:
                format_string = job.path_templates.compilation
            else:
                format_string = job.path_templates.soundtrack
        elif source.meta.comp:
            format_string = job.path_templates.compilation
        else:
            format_string = job.path_templates.default

        meta_dict = source.meta.export_dict()

        desired_target_path = process_target_path(
            meta_dict, format_string, job.template_settings.shell_friendly
        )

        # Remove the leading path separator to prevent the audio files from
        # ending up in a folder other than the target folder.
        desired_target_path = re.sub(r"^" + os.path.sep + r"+", "", desired_target_path)
        desired_target_path = os.path.join(
            job.selection.target, desired_target_path + "." + source.extension
        )

        desired_target = AudioFile(
            desired_target_path,
            job=job,
            prefix=job.selection.target,
            file_type="target",
        )

        # Do nothing
        if source.abspath == desired_target.abspath:
            job.msg.status("Renamed", status="ok")
            count("renamed")
            return

        # Search existing target
        target = False
        target_path = find_target_path(desired_target.abspath, job.filters.extension)
        if target_path:
            target = AudioFile(
                target_path, job=job, prefix=job.selection.target, file_type="target"
            )

        # Both file exist
        if target:
            if not target.meta:
                raise Exception("target.meta must not be empty.")
            best = detect_best_format(source.meta, target.meta, job)

            if job.rename.cleaning_action:
                # delete source
                if not job.rename.best_format or (
                    job.rename.best_format and best == "target"
                ):
                    action.cleanup(source)

                # delete target
                if job.rename.best_format and best == "source":
                    action.cleanup(target)

                    # Unset target object to trigger copy or move actions.
                    target = None

        if target:
            job.msg.status("Exists", status="error")

        # copy
        elif job.rename.move_action == "copy":
            action.copy(source, desired_target)

        # move
        elif job.rename.move_action == "move":
            action.move(source, desired_target)



"""Extend the class ``MediaFile`` of the package ``phrydy``."""

import re
from typing import Any, Dict, List, Optional, Tuple

from phrydy import MediaFileExtended
from tmep import Functions

import audiorename.musicbrainz as musicbrainz

Diff = List[Tuple[str, Optional[str], Optional[str]]]
PerformerRaw = List[List[str]]


def compare_dicts(first: Dict[str, str], second: Dict[str, str]) -> Diff:
    """Compare two dictionaries for differenes.

    :param first: First dictionary to diff.
    :param second: Second dicationary to diff.

    :return: As list of key entries whose values differ.
    """
    diff: Diff = []
    for key, _ in sorted(first.items()):
        if key not in second:
            diff.append((key, first[key], None))

    for key, _ in sorted(second.items()):
        if key not in first:
            diff.append((key, None, second[key]))

    all_keys: set[str] = set()
    for key, _ in first.items():
        all_keys.add(key)
    for key, _ in second.items():
        all_keys.add(key)
    for key in sorted(all_keys):
        if key in first and key in second and first[key] != second[key]:
            diff.append((key, first[key], second[key]))

    return diff


class Meta(MediaFileExtended):
    def __init__(self, path: str, shell_friendly: bool = False):
        super(Meta, self).__init__(path, False)
        self.shell_friendly = shell_friendly

    ###############################################################################
    # Public methods
    ###############################################################################

    def export_dict(self, sanitize: bool = True) -> Dict[str, str]:
        """
        Export all fields into a dictionary.

        :param sanitize: Set the parameter to true to trigger the sanitize
          function.
        """
        out: Dict[str, str] = {}
        for field in self.fields_sorted():
            value = getattr(self, field)
            if value:
                if sanitize:
                    out[field] = self._sanitize(str(value))
                else:
                    out[field] = value

        return out

    def enrich_metadata(self) -> None:
        musicbrainz.set_useragent()

        if self.mb_trackid:
            recording = musicbrainz.query("recording", self.mb_trackid)
        else:
            print("No music brainz track id found.")
            return

        release = None
        if self.mb_albumid:
            release = musicbrainz.query("release", self.mb_albumid)
        if release and "release-group" in release:
            release_group = release["release-group"]
            types: List[str] = []
            if "type" in release_group:
                types.append(release_group["type"])
            if "primary-type" in release_group:
                types.append(release_group["primary-type"])
            if "secondary-type-list" in release_group:
                types = types + release_group["secondary-type-list"]
            types = self._uniquify_list(types)
            self.releasegroup_types = "/".join(types).lower()

        work_id = ""
        if self.mb_workid:
            work_id = self.mb_workid
        elif recording:
            try:
                work_id = recording["work-relation-list"][0]["work"]["id"]
            except KeyError:
                pass
        if work_id:
            work_hierarchy = musicbrainz.query_works_recursively(work_id, [])
            if work_hierarchy:
                work_hierarchy.reverse()
                work_bottom = work_hierarchy[-1]
                if "artist-relation-list" in work_bottom:
                    for artist in work_bottom["artist-relation-list"]:
                        if (
                            artist["direction"] == "backward"
                            and artist["type"] == "composer"
                        ):
                            self.composer = artist["artist"]["name"]
                            self.composer_sort = artist["artist"]["sort-name"]
                            break
                self.mb_workid = work_bottom["id"]
                self.work = work_bottom["title"]
                wh_titles: List[str] = []
                wh_ids: List[str] = []
                for work in work_hierarchy:
                    wh_titles.append(work["title"])
                    wh_ids.append(work["id"])
                self.work_hierarchy = " -> ".join(wh_titles)
                self.mb_workhierarchy_ids = "/".join(wh_ids)

    def remap_classical(self) -> None:
        """Remap some fields to fit better for classical music:
        ``composer`` becomes ``artist``, ``work`` becomes ``album``, from the
        ``title`` the work prefix is removed (``Symphonie No. 9: I. Allegro``
        -> ``I. Allegro``) and ``track`` becomes the movement number. All
        overwritten fields are safed in the ``comments`` field. No combined
        properties (like ``ar_combined_composer``) are used and therefore some
        code duplications are done on purpose to avoid circular endless loops.
        """
        safe: List[List[str]] = []

        if self.title:
            safe.append(["title", self.title])
            self.title = re.sub(r"^[^:]*: ?", "", self.title)

            roman = re.findall(r"^([IVXLCDM]*)\.", self.title)
            if roman:
                safe.append(["track", str(self.track)])
                self.track = self._roman_to_int(roman[0])

        if self.composer:
            safe.append(["artist", self.artist])
            self.artist = self.composer

        if self.ar_combined_work_top:
            safe.append(["album", self.album])
            self.ar_performer_short
            album = self.ar_combined_work_top
            if self.ar_performer_short:
                album += " (" + self.ar_performer_short + ")"
            self.album = album

        if safe:
            comments = "Original metadata: "
            for safed in safe:
                comments = comments + str(safed[0]) + ": " + str(safed[1]) + "; "

            self.comments = comments

    ###############################################################################
    # Class methods
    ###############################################################################

    @classmethod
    def fields_phrydy(cls):
        for field in sorted(MediaFileExtended.readable_fields()):
            yield field

    @classmethod
    def fields_audiorename(cls):
        for prop, _ in sorted(cls.__dict__.items()):
            if isinstance(getattr(cls, prop), property):
                if isinstance(prop, bytes):
                    # On Python 2, class field names are bytes. This method
                    # produces text strings.
                    yield prop.decode("utf8", "ignore")
                else:
                    yield prop

    @classmethod
    def fields(cls):
        for field in cls.fields_phrydy():
            yield field
        for field in cls.fields_audiorename():
            yield field

    @classmethod
    def fields_sorted(cls):
        for field in sorted(cls.fields()):
            yield field

    ###############################################################################
    # Static methods
    ###############################################################################

    @staticmethod
    def _find_initials(value: str) -> str:
        """
        Find the first character of a string.

        :param str value: A string to extract the initials.

        :return: A single character in lowercase. The possible return values
            are lowercase letters from the ASCII alphabet (``a-z``), the digit
            ``0`` and the underscore character (``_``).
        """
        # To avoid ae -> a
        value = Functions.tmpl_asciify(value)
        # To avoid “!K7-Compilations” -> “!”
        value = re.sub(r"^\W*", "", value)
        initial = value[0:1].lower()

        if re.match(r"\d", initial):
            return "0"

        if initial == "":
            return "_"

        return initial

    @staticmethod
    def _normalize_performer(ar_performer: List[str]) -> PerformerRaw:
        """
        :param ar_performer: A list of raw ar_performer strings like

        .. code-block:: python

            ['John Lennon (vocals)', 'Ringo Starr (drums)']

        :return: A list

        .. code-block:: python

            [
                ['vocals', 'John Lennon'],
                ['drums', 'Ringo Starr'],
            ]
        """
        out: PerformerRaw = []
        for value in ar_performer:
            value = value[:-1]
            performers: List[str] = value.split(" (")
            if len(performers) == 2:
                out.append([performers[1], performers[0]])
        return out

    @staticmethod
    def _roman_to_int(n: str) -> int:
        numeral_map = tuple(
            zip(
                (1000, 900, 500, 400, 100, 90, 50, 40, 10, 9, 5, 4, 1),
                ("M", "CM", "D", "CD", "C", "XC", "L", "XL", "X", "IX", "V", "IV", "I"),
            )
        )
        i = result = 0
        for integer, numeral in numeral_map:
            while n[i : i + len(numeral)] == numeral:
                result += integer
                i += len(numeral)
        return result

    @staticmethod
    def _sanitize(value: Any) -> str:
        if isinstance(value, str) or isinstance(value, bytes):
            value = Functions.tmpl_sanitize(str(value))
            value = re.sub(r"\s{2,}", " ", str(value))
        else:
            value = ""
        return value

    @staticmethod
    def _shorten_performer(
        ar_performer: str,
        length: int = 3,
        separator: str = " ",
        abbreviation: str = ".",
    ) -> str:
        out = ""
        count = 0
        for s in ar_performer.split(" "):
            if count < 3:
                if len(s) > length:
                    part = s[:length] + abbreviation
                else:
                    part = s
                out = out + separator + part
            count = count + 1

        return out[len(separator) :]

    @staticmethod
    def _uniquify_list(sequence: List[Any]) -> List[Any]:
        """https://www.peterbe.com/plog/uniqifiers-benchmark"""
        unique: List[Any] = []

        for element in sequence:
            if not unique.count(element):
                unique.append(element)

        return unique

    ###############################################################################
    # Properties
    ###############################################################################

    @property
    def ar_classical_album(self) -> Optional[str]:
        """Uses:

        * ``phrydy.mediafile.MediaFile.work``

        Examples:

        * ``Horn Concerto: I. Allegro`` → ``Horn Concerto``
        * ``Die Meistersinger von Nürnberg``
        """
        if self.work:
            return re.sub(r":.*$", "", (str(self.work)))
        return None

    @property
    def ar_combined_album(self) -> Optional[str]:
        """Uses:

        * ``phrydy.mediafile.MediaFile.album``

        Example:

        * ``Just Friends (Disc 2)`` → ``Just Friends``
        """
        if self.album:
            return re.sub(r" ?\([dD]is[ck].*\)$", "", str(self.album))
        return None

    @property
    def ar_initial_album(self) -> Optional[str]:
        """Uses:

        * :class:`audiorename.meta.Meta.ar_combined_album`

        Examples:

        * ``Just Friends`` → ``j``
        * ``Die Meistersinger von Nürnberg``  → ``d``
        """
        if self.ar_combined_album:
            return self._find_initials(self.ar_combined_album)
        return None

    @property
    def ar_initial_artist(self) -> str:
        """Uses:

        * :class:`audiorename.meta.Meta.ar_combined_artist_sort`

        Examples:

        * ``Just Friends`` → ``j``
        * ``Die Meistersinger von Nürnberg``  → ``d``
        """
        return self._find_initials(self.ar_combined_artist_sort)

    @staticmethod
    def __remove_feat_vs_second_artist(artist: str) -> str:
        """Give only the first artist, remove the second after ``feat.``,
        ``ft.`` or ``vs.``"""
        return re.sub(r"\s+(feat|ft|vs)\.?\s.*", "", artist, flags=re.IGNORECASE)

    @property
    def ar_combined_artist(self) -> str:
        """Uses:

        * ``phrydy.mediafile.MediaFile.albumartist``
        * ``phrydy.mediafile.MediaFile.artist``
        * ``phrydy.mediafile.MediaFile.albumartist_credit``
        * ``phrydy.mediafile.MediaFile.artist_credit``
        * ``phrydy.mediafile.MediaFile.albumartist_sort``
        * ``phrydy.mediafile.MediaFile.artist_sort``

        Removes the second artist after ``feat.``, ``ft.`` or ``vs.``.
        """
        out: str
        if self.albumartist:
            out = self.albumartist
        elif self.artist:
            out = self.artist
        elif self.albumartist_credit:
            out = self.albumartist_credit
        elif self.artist_credit:
            out = self.artist_credit
        # Same as aristsafe_sort
        elif self.albumartist_sort:
            out = self.albumartist_sort
        elif self.artist_sort:
            out = self.artist_sort
        else:
            out = "Unknown"

        return Meta.__remove_feat_vs_second_artist(out)

    @property
    def ar_combined_artist_sort(self) -> str:
        """Uses:

        * ``phrydy.mediafile.MediaFile.albumartist_sort``
        * ``phrydy.mediafile.MediaFile.artist_sort``
        * ``phrydy.mediafile.MediaFile.albumartist``
        * ``phrydy.mediafile.MediaFile.artist``
        * ``phrydy.mediafile.MediaFile.albumartist_credit``
        * ``phrydy.mediafile.MediaFile.artist_credit``

        Removes the second artist after ``feat.``, ``ft.`` or ``vs.``.
        """
        out: str
        if self.albumartist_sort:
            out = self.albumartist_sort
        elif self.artist_sort:
            out = self.artist_sort
        # Same as ar_combined_artist
        elif self.albumartist:
            out = self.albumartist
        elif self.artist:
            out = self.artist
        elif self.albumartist_credit:
            out = self.albumartist_credit
        elif self.artist_credit:
            out = self.artist_credit
        else:
            out = "Unknown"

        out = Meta.__remove_feat_vs_second_artist(out)

        if self.shell_friendly:
            out = out.replace(", ", "_")

        return out

    @property
    def ar_initial_composer(self) -> str:
        """Uses:

        * :class:`audiorename.meta.Meta.ar_combined_composer`
        """
        return self._find_initials(self.ar_combined_composer)

    @property
    def ar_combined_composer(self) -> str:
        """Uses:

        * ``phrydy.mediafile.MediaFile.composer_sort``
        * ``phrydy.mediafile.MediaFile.composer``
        * :class:`audiorename.meta.Meta.ar_combined_artist`
        """
        out: str = ""
        if self.composer_sort:
            out = self.composer_sort
        elif self.composer:
            out = self.composer
        else:
            out = self.ar_combined_artist

        if self.shell_friendly:
            out = out.replace(", ", "_")

        # e. g. 'Mozart, Wolfgang Amadeus/Süßmeyer, Franz Xaver'
        return re.sub(r" ?/.*", "", out)

    @property
    def ar_combined_disctrack(self) -> Optional[str]:
        """
        Generate a combination of track and disc number, e. g.: ``1-04``,
        ``3-06``.

        Uses:

        * ``phrydy.mediafile.MediaFile.disctotal``
        * ``phrydy.mediafile.MediaFile.disc``
        * ``phrydy.mediafile.MediaFile.tracktotal``
        * ``phrydy.mediafile.MediaFile.track``
        """

        if not self.track:
            return None

        if self.disctotal and int(self.disctotal) > 99:
            disk = str(self.disc).zfill(3)
        elif self.disctotal and int(self.disctotal) > 9:
            disk = str(self.disc).zfill(2)
        else:
            disk = str(self.disc)

        if self.tracktotal and int(self.tracktotal) > 99:
            track = str(self.track).zfill(3)
        else:
            track = str(self.track).zfill(2)

        if self.disc and self.disctotal and int(self.disctotal) > 1:
            out = disk + "-" + track
        elif self.disc and not self.disctotal:
            out = disk + "-" + track
        else:
            out = track

        return out

    @property
    def ar_performer(self) -> str:
        """Uses:

        * :class:`audiorename.meta.Meta.ar_performer_raw`
        """
        out: str = ""
        for ar_performer in self.ar_performer_raw:
            out = out + ", " + ar_performer[1]

        out = out[2:]

        return out

    @property
    def ar_classical_performer(self) -> str:
        """http://musicbrainz.org/doc/Style/Classical/Release/Artist

        Uses:

        * :class:`audiorename.meta.Meta.ar_performer_short`
        * ``phrydy.mediafile.MediaFile.albumartist``
        """
        if len(self.ar_performer_short) > 0:
            out = self.ar_performer_short
        elif self.albumartist:
            out = re.sub(r"^.*; ?", "", self.albumartist)
        else:
            out = ""

        return out

    @property
    def ar_performer_raw(self) -> PerformerRaw:
        """Generate a unifed ar_performer list.

        Picard doesn’t store ar_performer values in m4a, alac.m4a, wma, wav,
        aiff.

        :return: A list

        .. code-block:: python

            [
                ['conductor', 'Herbert von Karajan'],
                ['violin', 'Anne-Sophie Mutter'],
            ]

        Uses:

        * ``phrydy.mediafile.MediaFile.mgfile``
        """
        out: PerformerRaw = []

        if (
            self.format == "FLAC" or self.format == "OGG"
        ) and "performer" in self.mgfile:
            out = self._normalize_performer(self.mgfile["performer"])
            if "conductor" in self.mgfile:
                out.insert(0, ["conductor", self.mgfile["conductor"][0]])
        elif self.format == "MP3":
            # 4.2.2 TMCL Musician credits list
            if "TMCL" in self.mgfile:
                out = self.mgfile["TMCL"].people
            # 4.2.2 TIPL Involved people list
            # TIPL is used for producer
            if "TIPL" in self.mgfile:
                out = self.mgfile["TIPL"].people

            # 4.2.2 TPE3 Conductor/ar_performer refinement
            if len(out) > 0 and "conductor" not in out[0] and "TPE3" in self.mgfile:
                out.insert(0, ["conductor", self.mgfile["TPE3"].text[0]])

        else:
            out = []

        return self._uniquify_list(out)

    @property
    def ar_performer_short(self):
        """Uses:

        * ``phrydy.mediafile.MediaFile.ar_performer_raw``
        """
        out: List[str] = []

        performers = self.ar_performer_raw
        picked: PerformerRaw = []
        for performer in performers:
            if performer[0] == "conductor" or performer[0] == "orchestra":
                picked.append(performer)

        if len(picked) > 0:
            performers = picked

        for performer in performers:
            if (
                performer[0] == "producer"
                or performer[0] == "executive producer"
                or performer[0] == "balance engineer"
            ):
                pass
            elif (
                performer[0] == "orchestra"
                or performer[0] == "choir vocals"
                or performer[0] == "string quartet"
            ):
                out.append(
                    self._shorten_performer(performer[1], separator="", abbreviation="")
                )
            else:
                out.append(performer[1].split(" ")[-1])

        return ", ".join(out)

    @property
    def ar_combined_soundtrack(self) -> bool:
        if self.releasegroup_types and "soundtrack" in self.releasegroup_types.lower():
            return True

        if self.albumtype and "soundtrack" in self.albumtype.lower():
            return True

        if self.albumtypes:
            for type in self.albumtypes:
                if "soundtrack" in type.lower():
                    return True

        return False

    @property
    def ar_classical_title(self) -> Optional[str]:
        """Uses:

        * ``phrydy.mediafile.MediaFile.title``

        Example:

        * ``Horn Concerto: I. Allegro``
        """
        if self.title:
            return re.sub(r"^[^:]*: ?", "", self.title)
        return None

    @property
    def ar_classical_track(self) -> Optional[str]:
        """Uses:

        * :class:`audiorename.meta.Meta.ar_classical_title`
        * :class:`audiorename.meta.Meta.ar_combined_disctrack`
        """
        roman = None
        if self.ar_classical_title:
            roman = re.findall(r"^([IVXLCDM]*)\.", self.ar_classical_title)
        if roman:
            return str(self._roman_to_int(roman[0])).zfill(2)
        elif self.ar_combined_disctrack:
            return self.ar_combined_disctrack
        return None

    @property
    def ar_combined_work_top(self) -> Optional[str]:
        """Uses:

        * ``phrydy.mediafile.MediaFile.work_hierarchy``
        * ``phrydy.mediafile.MediaFile.work``

        """
        if self.work_hierarchy:
            return self.work_hierarchy.split(" -> ")[0]
        elif self.ar_classical_album:
            return self.ar_classical_album
        return None

    @property
    def ar_combined_year(self):
        """Uses:

        * ``phrydy.mediafile.MediaFile.original_year``
        * ``phrydy.mediafile.MediaFile.year``
        """
        if self.original_year:
            return self.original_year
        elif self.year:
            return self.year



"""Print messages on the command line."""

import typing
from collections import OrderedDict

import ansicolor
import phrydy
import tmep

import audiorename

from .args import all_fields

if typing.TYPE_CHECKING:
    from .audiofile import AudioFile
    from .job import Job
    from .meta import Meta


class KeyValue:
    kv: "OrderedDict[str, str]"

    def __init__(self, color: bool = False):
        self.color = color
        self.kv = OrderedDict()

    def add(self, key: str, value: str) -> None:
        self.kv[key] = value

    def result(self) -> str:
        out = ""
        for key, value in self.kv.items():
            key = key + ":"
            if self.color:
                key = ansicolor.yellow(key)
            out = out + key + " " + value + "\n"
        return out

    def result_one_line(self) -> str:
        out: typing.List[str] = []
        for key, value in self.kv.items():
            if self.color:
                key = ansicolor.green(key)
            out.append(key + "=" + str(value))

        return " ".join(out)


class Message:
    """Print messages on the command line interface.

    :param job: The `job` object.
    :type job: audiorename.job.Job
    """

    color: bool
    verbose: bool
    one_line: bool

    def __init__(self, job: "Job"):
        self.color = job.cli_output.color
        self.verbose = job.cli_output.verbose
        self.one_line = job.cli_output.one_line
        self.max_field = self.max_fields_length()
        self.indent_width = 4

    @staticmethod
    def max_fields_length() -> int:
        return phrydy.doc_generator.get_max_field_length(all_fields)

    def output(self, text: str = "") -> None:
        if self.one_line:
            print(text.strip(), end=" ")
        else:
            print(text)

    def template_indent(self, level: int = 1) -> str:
        return (" " * self.indent_width) * level

    def template_path(self, audio_file: "AudioFile") -> str:
        path: str
        if self.verbose:
            path = audio_file.abspath
        else:
            path = audio_file.short

        if self.color:
            if audio_file.type == "source":
                path = ansicolor.magenta(path)
            else:
                path = ansicolor.yellow(path)

        return path

    def next_file(self, audio_file: "AudioFile"):
        print()
        if self.verbose:
            path = audio_file.abspath
        else:
            path = audio_file.dir_and_file
        if self.color:
            path = ansicolor.blue(path, reverse=True)
        self.output(path)

    def action_one_path(self, message: str, audio_file: "AudioFile") -> None:
        self.status(message, status="progress")
        self.output(self.template_indent(2) + self.template_path(audio_file))
        self.output()

    def action_two_path(
        self, message: str, source: "AudioFile", target: "AudioFile"
    ) -> None:
        self.status(message, status="progress")
        self.output(self.template_indent(2) + self.template_path(source))
        self.output(self.template_indent(2) + "to:")
        self.output(self.template_indent(2) + self.template_path(target))

    def best_format(
        self,
        best: typing.Literal["target", "source"],
        attr: str,
        source: "Meta",
        target: "Meta",
    ) -> None:
        source_attr = getattr(source, attr)
        target_attr = getattr(target, attr)

        if source_attr == target_attr and best == "target":
            self.output(
                "Best format: Source and target have the same formats," " use target."
            )
        else:
            self.output(
                "Best format is “" + best + "” because of “" + attr + "”: "
                "(source: " + str(source_attr) + ", "
                "target: " + str(target_attr) + ")"
            )

    def diff(self, key: str, value1: str, value2: str):
        key_width = self.max_field + 2
        value2_indent = self.indent_width + key_width
        key += ":"
        key = key.ljust(self.max_field + 2)

        def quote(value: typing.Any):
            return "“" + str(value) + "”"

        value1 = quote(value1)
        value2 = quote(value2)
        if self.color:
            value1 = ansicolor.red(value1)
            value2 = ansicolor.green(value2)
            key = ansicolor.yellow(key)
        self.output(self.template_indent(1) + key + value1)
        self.output(" " * value2_indent + value2)

    def status_color(
        self, status: typing.Literal["ok", "error", "progress"]
    ) -> typing.Literal["yellow", "red", "green"]:
        if status == "progress":
            return "yellow"
        elif status == "error":
            return "red"
        else:
            return "green"

    def status(self, text: str, status: typing.Literal["ok", "error", "progress"]):
        if self.color:
            color = getattr(ansicolor, self.status_color(status))
            text = color(text, reverse=True)
        self.output(self.template_indent(1) + text)


def job_info(job: "Job") -> None:
    versions = KeyValue(job.cli_output.color)
    versions.add("audiorename", audiorename.__version__)
    versions.add("phrydy", phrydy.__version__)
    versions.add("tmep", tmep.__version__)

    info = KeyValue(job.cli_output.color)
    info.add("Versions", versions.result_one_line())
    info.add("Action", job.rename.move_action)
    info.add("Source", job.selection.source)
    if job.selection.target:
        info.add("Target", job.selection.target)
    if job.rename.cleaning_action == "backup":
        info.add("Backup folder", job.rename.backup_folder)

    if job.cli_output.verbose:
        info.add("Default", job.path_templates.default)
        info.add("Compilation", job.path_templates.compilation)
        info.add("Soundtrack", job.path_templates.soundtrack)

    print(info.result())


def stats(job: "Job") -> None:
    kv = KeyValue(job.cli_output.color)

    kv.add("Execution time", job.stats.timer.result())
    kv.add("Counter", job.stats.counter.result())
    print(kv.result())



"""Collect all informations about the current job in a class."""

import configparser
import os
import time
import typing

from .args import ArgsDefault
from .message import Message


class Timer:
    begin: float = 0

    end: float = 0

    def start(self) -> None:
        self.begin = time.time()

    def stop(self) -> None:
        self.end = time.time()

    def result(self) -> str:
        return "{:.1f}s".format(self.end - self.begin)


class Counter:
    def __init__(self) -> None:
        self._counters: typing.Dict[str, int] = {}

    def reset(self) -> None:
        self._counters = {}

    def count(self, counter: str) -> None:
        """Add one to number identified by a string.

        :param counter: A string to identify the counter

        :return: None
        """
        if counter in self._counters:
            self._counters[counter] += 1
        else:
            self._counters[counter] = 1

    def get(self, counter: str) -> int:
        """Get the counter identify by a string.

        :param str counter: A string to identify the counter

        :return: The counter as a number
        :rtype: int
        """

        if counter in self._counters:
            return self._counters[counter]
        else:
            return 0

    def result(self) -> str:
        out: typing.List[str] = []
        for counter, value in sorted(self._counters.items()):
            out.append(counter + "=" + str(value))

        if out:
            return " ".join(out)
        else:
            return "Nothing to count!"


class Statistic:
    counter = Counter()
    timer = Timer()


IniDataTypes = typing.Literal["boolean", "integer", "string"]


class Config:
    """The class ``Config`` is used to combine the two sources of settings
    (command line arguments and INI configuration file). The command line
    arguments override the values of the configuration file. This class is to
    be inherited by subclasses. Each subclass corresponds to a section of the
    INI configuration file. All settings are saved as private properties with
    leading underscore. The subclass provide  for each private property a
    getter method (@property)"""

    _job: "Job"

    def __init__(
        self, job: "Job", section: str, options: typing.Dict[str, IniDataTypes]
    ):
        self._job = job
        for key, data_type in options.items():
            value = None
            if getattr(job.args, key) is not None:
                value = getattr(job.args, key)
            elif job.config:
                for config in job.config:
                    result = self.__get_value_from_config(
                        config, section, key, data_type
                    )
                    if result is not None:
                        value = result

            if value is not None:
                setattr(self, "_" + key, value)

    def __get_value_from_config(
        self,
        config: configparser.ConfigParser,
        section: str,
        key: str,
        data_type: IniDataTypes,
    ) -> typing.Optional[typing.Any]:
        try:
            if data_type == "boolean":
                return config.getboolean(section, key)
            elif data_type == "integer":
                return config.getint(section, key)
            else:
                return config.get(section, key)
        except (configparser.NoOptionError, configparser.NoSectionError):
            pass
        return None


class SelectionConfig(Config):
    _source: typing.Optional[str]
    _target: typing.Optional[str]
    _source_as_target: typing.Optional[bool]

    @property
    def source(self) -> str:
        """The source path as an absolute path. It maybe a directory or a
        file."""
        source: str
        if hasattr(self, "_source") and self._source:
            source = self._source
        else:
            source = "."
        return os.path.abspath(source)

    @property
    def target(self) -> typing.Optional[str]:
        """The path of the target as an absolute path. It is always a
        directory.
        """
        target: typing.Union[None, str] = None
        if hasattr(self, "_target"):
            target = self._target

        if self.source_as_target:
            if os.path.isdir(self.source):
                return os.path.abspath(self.source)
            else:
                return os.path.abspath(os.path.dirname(self.source))
        elif target:
            return os.path.abspath(target)
        else:
            return os.getcwd()

    @property
    def source_as_target(self) -> bool:
        if hasattr(self, "_source_as_target") and isinstance(
            self._source_as_target, bool
        ):
            return self._source_as_target
        return False


MoveAction = typing.Literal["move", "copy", "no_rename"]
CleaningAction = typing.Literal["backup", "delete", "do_nothing"]


class RenameConfig(Config):
    _backup_folder: typing.Optional[str]
    _best_format: typing.Optional[bool]
    _dry_run: typing.Optional[bool]
    _move_action: typing.Optional[MoveAction]
    _cleaning_action: typing.Optional[CleaningAction]

    @property
    def backup_folder(self) -> str:
        if hasattr(self, "_backup_folder") and isinstance(self._backup_folder, str):
            return self._backup_folder
        return os.path.join(os.getcwd(), "_audiorename_backups")

    @property
    def best_format(self) -> bool:
        if hasattr(self, "_best_format") and isinstance(self._best_format, bool):
            return self._best_format
        return True

    @property
    def dry_run(self) -> bool:
        if hasattr(self, "_dry_run") and isinstance(self._dry_run, bool):
            return self._dry_run
        return False

    @property
    def move_action(self) -> MoveAction:
        if hasattr(self, "_move_action") and self._move_action in [
            "move",
            "copy",
            "no_rename",
        ]:
            return self._move_action
        return "move"

    @property
    def cleaning_action(self) -> CleaningAction:
        if hasattr(self, "_cleaning_action") and self._cleaning_action in [
            "backup",
            "delete",
            "do_nothing",
        ]:
            return self._cleaning_action
        return "do_nothing"


class FiltersConfig(Config):
    _album_complete: typing.Optional[bool]
    _album_min: typing.Optional[int]
    _extension: typing.Optional[str]
    _genre_classical: typing.Optional[str]
    _field_skip: typing.Optional[str]

    @property
    def album_complete(self) -> bool:
        if hasattr(self, "_album_complete") and isinstance(self._album_complete, bool):
            return self._album_complete
        return False

    @property
    def album_min(self) -> typing.Optional[int]:
        if hasattr(self, "_album_min") and isinstance(self._album_min, int):
            return self._album_min
        return None

    @property
    def extension(self) -> typing.List[str]:
        extension: str
        if hasattr(self, "_extension") and isinstance(self._extension, str):
            extension = self._extension
        else:
            extension = "mp3,m4a,flac,wma"
        return extension.split(",")

    @property
    def genre_classical(self) -> typing.List[str]:
        genre_classical: str
        if hasattr(self, "_genre_classical") and isinstance(self._genre_classical, str):
            genre_classical = self._genre_classical
        else:
            genre_classical = ","
        return list(filter(str.strip, genre_classical.lower().split(",")))

    @property
    def field_skip(self) -> typing.Optional[str]:
        if hasattr(self, "_field_skip") and isinstance(self._field_skip, str):
            return self._field_skip
        return None


class TemplateSettingsConfig(Config):
    _classical: typing.Optional[bool]
    _shell_friendly: typing.Optional[bool]
    _no_soundtrack: typing.Optional[bool]

    @property
    def classical(self) -> bool:
        if hasattr(self, "_classical") and isinstance(self._classical, bool):
            return self._classical
        return False

    @property
    def shell_friendly(self) -> bool:
        if hasattr(self, "_shell_friendly") and isinstance(self._shell_friendly, bool):
            return self._shell_friendly
        return False

    @property
    def no_soundtrack(self) -> bool:
        if hasattr(self, "_no_soundtrack") and isinstance(self._no_soundtrack, bool):
            return self._no_soundtrack
        return False


class PathTemplatesConfig(Config):
    """A class to store the selected or configured path templates. This class
    can be accessed under the attibute path_templates of the Job class."""

    _default_template: typing.Optional[str]
    _compilation_template: typing.Optional[str]
    _soundtrack_template: typing.Optional[str]
    _classical_template: typing.Optional[str]

    @property
    def _is_classical(self) -> bool:
        return self._job.template_settings.classical

    @property
    def default(self) -> str:
        """Get the default path template."""
        if self._is_classical:
            return self.classical
        if hasattr(self, "_default_template") and isinstance(
            self._default_template, str
        ):
            return self._default_template
        return (
            "$ar_initial_artist/"
            "%shorten{$ar_combined_artist_sort}/"
            "%shorten{$ar_combined_album}"
            "%ifdefnotempty{ar_combined_year,_${ar_combined_year}}/"
            "${ar_combined_disctrack}_%shorten{$title}"
        )

    @property
    def compilation(self) -> str:
        """Get the path template for compilations."""
        if self._is_classical:
            return self.classical
        if hasattr(self, "_compilation_template") and isinstance(
            self._compilation_template, str
        ):
            return self._compilation_template
        return (
            "_compilations/"
            "$ar_initial_album/"
            "%shorten{$ar_combined_album}"
            "%ifdefnotempty{ar_combined_year,_${ar_combined_year}}/"
            "${ar_combined_disctrack}_%shorten{$title}"
        )

    @property
    def soundtrack(self) -> str:
        """Get the path template for soundtracks."""
        if self._is_classical:
            return self.classical
        if self._job.template_settings.no_soundtrack:
            return self.default
        if hasattr(self, "_soundtrack_template") and isinstance(
            self._soundtrack_template, str
        ):
            return self._soundtrack_template
        return (
            "_soundtrack/"
            "$ar_initial_album/"
            "%shorten{$ar_combined_album}"
            "%ifdefnotempty{ar_combined_year,_${ar_combined_year}}/"
            "${ar_combined_disctrack}_${artist}_%shorten{$title}"
        )

    @property
    def classical(self) -> str:
        """Get the path template for classical music."""
        if hasattr(self, "_classical_template") and isinstance(
            self._classical_template, str
        ):
            return self._classical_template
        return (
            "$ar_initial_composer/$ar_combined_composer/"
            "%shorten{$ar_combined_work_top,48}"
            "_[%shorten{$ar_classical_performer,32}]/"
            "${ar_combined_disctrack}_%shorten{$ar_classical_title,64}"
            "%ifdefnotempty{acoustid_id,_%shorten{$acoustid_id,8}}"
        )


class CliOutputConfig(Config):
    _color: typing.Optional[bool]
    _debug: typing.Optional[bool]
    _job_info: typing.Optional[bool]
    _mb_track_listing: typing.Optional[bool]
    _one_line: typing.Optional[bool]
    _stats: typing.Optional[bool]
    _verbose: typing.Optional[bool]

    @property
    def color(self) -> bool:
        if hasattr(self, "_color") and isinstance(self._color, bool):
            return self._color
        return True

    @property
    def debug(self) -> bool:
        if hasattr(self, "_debug") and isinstance(self._debug, bool):
            return self._debug
        return False

    @property
    def job_info(self) -> bool:
        if hasattr(self, "_job_info") and isinstance(self._job_info, bool):
            return self._job_info
        return False

    @property
    def mb_track_listing(self) -> bool:
        if hasattr(self, "_mb_track_listing") and isinstance(
            self._mb_track_listing, bool
        ):
            return self._mb_track_listing
        return False

    @property
    def one_line(self) -> bool:
        if hasattr(self, "_one_line") and isinstance(self._one_line, bool):
            return self._one_line
        return False

    @property
    def stats(self) -> bool:
        if hasattr(self, "_stats") and isinstance(self._stats, bool):
            return self._stats
        return False

    @property
    def verbose(self) -> bool:
        if hasattr(self, "_verbose") and isinstance(self._verbose, bool):
            return self._verbose
        return False


class MetadataActionsConfig(Config):
    _enrich_metadata: typing.Optional[bool]
    _remap_classical: typing.Optional[bool]

    @property
    def enrich_metadata(self) -> bool:
        if hasattr(self, "_enrich_metadata") and isinstance(
            self._enrich_metadata, bool
        ):
            return self._enrich_metadata
        return False

    @property
    def remap_classical(self) -> bool:
        if hasattr(self, "_remap_classical") and isinstance(
            self._remap_classical, bool
        ):
            return self._remap_classical
        return False


class Job:
    """Holds informations of one job which can handle multiple files.

    A jobs represents one call of the program on the command line. This class
    unifies and processes the data of the `argparse` and the `configparser`
    call. It groups the `argparse` and the `configparser` key-value pairs into
    parent properties. The properties of this class for example can be used to
    display easily an overview message of the job.
    """

    stats = Statistic()

    args: ArgsDefault
    config: typing.Optional[typing.List[configparser.ConfigParser]] = None

    def __init__(self, args: ArgsDefault):
        self.args = args
        if args.config is not None:
            self.config = self.__read_config(args.config)

        self.msg = Message(self)

    def __read_config(
        self, file_paths: typing.List[str]
    ) -> typing.List[configparser.ConfigParser]:
        configs: typing.List[configparser.ConfigParser] = []
        for file_path in file_paths:
            config = configparser.ConfigParser()
            config.read(file_path)
            configs.append(config)
        return configs

    @property
    def selection(self) -> SelectionConfig:
        return SelectionConfig(
            self,
            "selection",
            {"source": "string", "target": "string", "source_as_target": "boolean"},
        )

    @property
    def rename(self) -> RenameConfig:
        return RenameConfig(
            self,
            "rename",
            {
                "backup_folder": "string",
                "best_format": "boolean",
                "dry_run": "boolean",
                "move_action": "string",
                "cleaning_action": "string",
            },
        )

    @property
    def filters(self) -> FiltersConfig:
        return FiltersConfig(
            self,
            "filters",
            {
                "album_complete": "boolean",
                "album_min": "integer",
                "extension": "string",
                "genre_classical": "string",
                "field_skip": "string",
            },
        )

    @property
    def template_settings(self) -> TemplateSettingsConfig:
        return TemplateSettingsConfig(
            self,
            "template_settings",
            {
                "classical": "boolean",
                "shell_friendly": "boolean",
                "no_soundtrack": "boolean",
            },
        )

    @property
    def path_templates(self) -> PathTemplatesConfig:
        return PathTemplatesConfig(
            self,
            "path_templates",
            {
                "default_template": "string",
                "compilation_template": "string",
                "soundtrack_template": "string",
                "classical_template": "string",
            },
        )

    @property
    def cli_output(self) -> CliOutputConfig:
        return CliOutputConfig(
            self,
            "cli_output",
            {
                "color": "boolean",
                "debug": "boolean",
                "job_info": "boolean",
                "mb_track_listing": "boolean",
                "one_line": "boolean",
                "stats": "boolean",
                "verbose": "boolean",
            },
        )

    @property
    def metadata_actions(self) -> MetadataActionsConfig:
        return MetadataActionsConfig(
            self,
            "metadata_actions",
            {
                "enrich_metadata": "boolean",
                "remap_classical": "boolean",
            },
        )



import typing
from typing import List, Literal, TypedDict, cast

import musicbrainzngs as musicbrainz

"""Query the musicbrainz API using the library
`musicbrainzngs <https://pypi.org/project/musicbrainzngs>`_.

.. code-block:: Python

    import json
    print(json.dumps(result,indent=2))

``get_recording_by_id`` with ``work-rels``

soundtrack/Pulp-Fiction/01.mp3

.. code-block:: JSON

    {
      "recording": {
        "length": "149000",
        "id": "0480672d-4d88-4824-a06b-917ff408eabe",
        "title": "Pumpkin and Honey Bunny ..."
      }
    }

classical/Mozart_Horn-concertos/01.mp3

.. code-block:: JSON

    {
      "recording": {
        "length": "286826",
        "work-relation-list": [
          {
            "type-id": "a3005666-a872-32c3-ad06-98af558e99b0",
            "begin": "1987-03",
            "end": "1987-03",
            "target": "21fe0bf0-a040-387c-a39d-369d53c251fe",
            "ended": "true",
            "work": {
              "id": "21fe0bf0-a040-387c-a39d-369d53c251fe",
              "language": "zxx",
              "title": "Concerto [...] KV 412: I. Allegro"
            },
            "type": "performance"
          }
        ],
        "id": "7886ad6c-11af-435b-8ec3-bca5711f7728",
        "title": "Konzert f\u00fcr [...] K. 386b/514: I. Allegro"
      }
    }


``get_release_by_id`` with ``release-groups``

soundtrack/Pulp-Fiction/01.mp3

.. code-block:: JSON

    {
      "release": {
        "status": "Bootleg",
        "release-event-count": 1,
        "title": "Pulp Fiction",
        "country": "US",
        "cover-art-archive": {
          "count": "1",
          "front": "true",
          "back": "false",
          "artwork": "true"
        },
        "release-event-list": [
          {
            "date": "2005-12-01",
            "area": {
              "sort-name": "United States",
              "iso-3166-1-code-list": [
                "US"
              ],
              "id": "489ce91b-6658-3307-9877-795b68554c98",
              "name": "United States"
            }
          }
        ],
        "release-group": {
          "first-release-date": "1994-09-27",
          "secondary-type-list": [
            "Compilation",
            "Soundtrack"
          ],
          "primary-type": "Album",
          "title": "Pulp Fiction: Music From the Motion Picture",
          "type": "Soundtrack",
          "id": "1703cd63-9401-33c0-87c6-50c4ba2e0ba8"
        },
        "text-representation": {
          "language": "eng",
          "script": "Latn"
        },
        "date": "2005-12-01",
        "quality": "normal",
        "id": "ab81edcb-9525-47cd-8247-db4fa969f525",
        "asin": "B000002OTL"
      }
    }

classical/Mozart_Horn-concertos/01.mp3

.. code-block:: JSON

    {
      "release": {
        "status": "Official",
        "release-event-count": 1,
        "title": "4 Hornkonzerte (Concertos for Horn and Orchestra)",
        "country": "DE",
        "barcode": "028942781429",
        "cover-art-archive": {
          "count": "0",
          "front": "false",
          "back": "false",
          "artwork": "false"
        },
        "release-event-list": [
          {
            "date": "1988",
            "area": {
              "sort-name": "Germany",
              "iso-3166-1-code-list": [
                "DE"
              ],
              "id": "85752fda-13c4-31a3-bee5-0e5cb1f51dad",
              "name": "Germany"
            }
          }
        ],
        "release-group": {
          "first-release-date": "1988",
          "title": "4 Hornkonzerte (Concertos for Horn and Orchestra)",
          "type": "Album",
          "id": "e1fa28f0-e56e-395b-82d3-a8de54e8c627",
          "primary-type": "Album"
        },
        "text-representation": {
          "language": "deu",
          "script": "Latn"
        },
        "date": "1988",
        "quality": "normal",
        "id": "5ed650c5-0f72-4b79-80a7-c458c869f53e",
        "asin": "B00000E4FA"
      }
    }

"""

WorkChild = TypedDict(
    "Work",
    {
        "id": str,
        "language": str,
        "title": str,
    },
)

WorkRelation = TypedDict(
    "WorkRelation",
    {"work": WorkChild, "direction": Literal["backward"], "type": Literal["parts"]},
)

Artist = TypedDict("Artist", {"name": str, "sort-name": str})

ArtistRelation = TypedDict(
    "ArtistRelation",
    {"artist": Artist, "direction": Literal["backward"], "type": Literal["composer"]},
)

Work = TypedDict(
    "Work",
    {
        "id": str,
        "language": str,
        "title": str,
        "work-relation-list": List[WorkRelation],
        "artist-relation-list": List[ArtistRelation],
    },
)
"""
``get_work_by_id`` with ``work-rels``

.. code-block:: JSON

    {
      "work": {
        "work-relation-list": [
          {
            "type-id": "ca8d3642-ce5f-49f8-91f2-125d72524e6a",
            "direction": "backward",
            "target": "5adc213f-700a-4435-9e95-831ed720f348",
            "ordering-key": "3",
            "work": {
              "id": "5adc213f-700a-4435-9e95-831ed720f348",
              "language": "deu",
              "title": "Die Zauberfl\u00f6te, K. 620: Akt I"
            },
            "type": "parts"
          },
          {
            "type-id": "51975ed8-bbfa-486b-9f28-5947f4370299",
            "work": {
              "disambiguation": "for piano, arr. Matthias",
              "id": "798f4c25-0ab3-44ba-81b6-3d856aedf82a",
              "language": "zxx",
              "title": "Die Zauberfl\u00f6te, K. 620: Aria ..."
            },
            "type": "arrangement",
            "target": "798f4c25-0ab3-44ba-81b6-3d856aedf82a"
          }
        ],
        "type": "Aria",
        "id": "eafec51f-47c5-3c66-8c36-a524246c85f8",
        "language": "deu",
        "title": "Die Zauberfl\u00f6te: Act I, Scene II. No. 2 Aria ..",
        "artist-relation-list": [
          {
            "type-id": "7474ab81-486f-40b5-8685-3a4f8ea624cb",
            "direction": "backward",
            "type": "librettist",
            "target": "86104c7c-cda4-4798-a4ab-104318c7ae9c",
            "artist": {
              "sort-name": "Schikaneder, Emanuel",
              "id": "86104c7c-cda4-4798-a4ab-104318c7ae9c",
              "name": "Emanuel Schikaneder"
            }
          },
          {
            "begin": "1791",
            "end": "1791",
            "target": "b972f589-fb0e-474e-b64a-803b0364fa75",
            "artist": {
              "sort-name": "Mozart, Wolfgang Amadeus",
              "disambiguation": "classical composer",
              "id": "b972f589-fb0e-474e-b64a-803b0364fa75",
              "name": "Wolfgang Amadeus Mozart"
            },
            "direction": "backward",
            "type-id": "d59d99ea-23d4-4a80-b066-edca32ee158f",
            "ended": "true",
            "type": "composer"
          }
        ]
      }
    }


.. code-block:: JSON

    {
      "work": {
        "work-relation-list": [
          {
            "type-id": "c1dca2cd-194c-36dd-93f8-6a359167e992",
            "direction": "backward",
            "work": {
              "id": "70e53569-258c-463d-9505-5b69dcbf374a",
              "title": "Can\u2019t Stop the Classics, Part 2"
            },
            "type": "medley",
            "target": "70e53569-258c-463d-9505-5b69dcbf374a"
          },
          {
            "type-id": "ca8d3642-ce5f-49f8-91f2-125d72524e6a",
            "direction": "backward",
            "target": "73663bd3-392f-45a7-b4ff-e75c01f5926a",
            "ordering-key": "1",
            "work": {
              "id": "73663bd3-392f-45a7-b4ff-e75c01f5926a",
              "language": "deu",
              "title": "Die Meistersinger von N\u00fcrnberg, WWV 96: Akt I"
            },
            "type": "parts"
          }
        ]
      }
    }
"""


def set_useragent() -> None:
    musicbrainz.set_useragent(
        "audiorename",
        "0.0.0",
        "https://github.com/Josef-Friedrich/audiorename",
    )


def query(
    mb_type: typing.Literal["recording", "work", "release"], mb_id: str
) -> typing.Union[typing.Dict[str, typing.Any], None]:
    method = "get_" + mb_type + "_by_id"
    query = getattr(musicbrainz, method)

    if mb_type == "recording" or mb_type == "work":
        mb_includes = ["work-rels"]
    elif mb_type == "release":
        mb_includes = ["release-groups"]
    else:
        mb_includes = []

    if mb_type == "work":
        mb_includes.append("artist-rels")

    try:
        result = query(mb_id, includes=mb_includes)
        return result[mb_type]

    except musicbrainz.ResponseError as err:
        if err.cause and err.cause.code == 404:
            print(
                "Item of type “" + mb_type + "” with the ID "
                "“" + mb_id + "” not found."
            )
        else:
            print("Received bad response from the MusicBrainz server.")
    return None


def query_works_recursively(work_id: str, works: List[Work] = []) -> List[Work]:
    work = cast(Work, query("work", work_id))

    if not work:
        return works

    works.append(work)

    parent_work = False
    if "work-relation-list" in work:
        for relation in work["work-relation-list"]:
            if (
                "direction" in relation
                and relation["direction"] == "backward"
                and relation["type"] == "parts"
            ):
                parent_work = relation
                break

    if parent_work:
        query_works_recursively(parent_work["work"]["id"], works)

    return works
